@@ -151,16 +151,21 @@ static int get_transport_payload(mpeg3_demuxer_t *demuxer,
 
 	if(bytes < 0)
 	{
-		printf("get_transport_payload: got negative payload size!\n");
+		fprintf(stderr, "get_transport_payload: got negative payload size!\n");
 		return 1;
 	}
 /*
  * 	if(demuxer->data_size + bytes > MPEG3_RAW_SIZE)
  * 		bytes = MPEG3_RAW_SIZE - demuxer->data_size;
  */
+//printf("get_transport_payload 2 %d %d %d\n", bytes, demuxer->read_all, is_audio);
 
 	if(demuxer->read_all && is_audio)
 	{
+/*
+ * if(demuxer->pid == 0x1100) 
+ * printf("get_transport_payload 1 0x%x %d\n", demuxer->audio_pid, bytes);
+ */
 		memcpy(demuxer->audio_buffer + demuxer->audio_size,
 			demuxer->raw_data + demuxer->raw_offset,
 			bytes);
@@ -169,6 +174,7 @@ static int get_transport_payload(mpeg3_demuxer_t *demuxer,
 	else
 	if(demuxer->read_all && is_video)
 	{
+//printf("get_transport_payload 2\n");
 		memcpy(demuxer->video_buffer + demuxer->video_size,
 			demuxer->raw_data + demuxer->raw_offset,
 			bytes);
@@ -180,6 +186,12 @@ static int get_transport_payload(mpeg3_demuxer_t *demuxer,
 			demuxer->raw_data + demuxer->raw_offset,
 			bytes);
 		demuxer->data_size += bytes;
+/*
+ * printf("get_transport_payload 10 pid=0x%x bytes=0x%x data_size=0x%x\n", 
+ * demuxer->pid, 
+ * bytes,
+ * demuxer->data_size);
+ */
 	}
 
 	demuxer->raw_offset += bytes;
@@ -255,16 +267,18 @@ static int get_unknown_data(mpeg3_demuxer_t *demuxer)
 
 
 
-static int get_pes_packet_data(mpeg3_demuxer_t *demuxer)
+static int get_transport_pes_packet(mpeg3_demuxer_t *demuxer)
 {
 	unsigned int pts = 0, dts = 0;
 	get_pes_packet_header(demuxer, &pts, &dts);
 
 
-	if(demuxer->stream_id == 0xbd)
+	if(demuxer->stream_id == 0xbd ||
+// Blu-Ray
+		demuxer->stream_id == 0xfd)
 	{
+//printf("get_transport_pes_packet %x\n", demuxer->pid);
 // AC3 audio
-// Don't know if the next byte is the true stream id like in program stream
 		demuxer->stream_id = 0x0;
 		demuxer->got_audio = 1;
 		demuxer->custom_id = demuxer->pid;
@@ -276,7 +290,12 @@ static int get_pes_packet_data(mpeg3_demuxer_t *demuxer)
 
 		if(demuxer->dump)
 		{
-			printf(" 0x%x bytes AC3 audio\n", demuxer->raw_size - demuxer->raw_offset);
+			printf("get_transport_pes_packet: offset=%llx 0x%x bytes AC3 custom_id=0x%x astream=0x%x do_audio=%d\n", 
+				mpeg3io_tell(demuxer->titles[demuxer->current_title]->fs),
+				demuxer->raw_size - demuxer->raw_offset,
+				demuxer->custom_id,
+				demuxer->astream,
+				demuxer->do_audio);
 		}
 
     	if((demuxer->custom_id == demuxer->astream && 
@@ -344,15 +363,6 @@ static int get_pes_packet_data(mpeg3_demuxer_t *demuxer)
 		{
 			demuxer->pes_video_time = (double)pts / 90000;
 			demuxer->video_pid = demuxer->pid;
-/*
- * printf("get_pes_packet_data video %04x %llx\n", 
- * stream_id, 
- * mpeg3io_tell(demuxer->titles[demuxer->current_title]->fs));
- * int i;
- * for(i = demuxer->raw_offset; i < demuxer->raw_size; i++)
- * printf("%02x ", demuxer->raw_data[i], stdout);
- * printf("\n");
- */
 
 
 			return get_transport_payload(demuxer, 0, 1);
@@ -386,7 +396,7 @@ static int get_pes_packet(mpeg3_demuxer_t *demuxer)
 	if(demuxer->stream_id != MPEG3_PRIVATE_STREAM_2 && 
 		demuxer->stream_id != MPEG3_PADDING_STREAM)
 	{
-		return get_pes_packet_data(demuxer);
+		return get_transport_pes_packet(demuxer);
 	}
 	else
 	if(demuxer->stream_id == MPEG3_PRIVATE_STREAM_2)
@@ -412,7 +422,7 @@ static int get_pes_packet(mpeg3_demuxer_t *demuxer)
 
 static int get_payload(mpeg3_demuxer_t *demuxer)
 {
-//printf("get_payload 1 %x %d\n", demuxer->pid, demuxer->payload_unit_start_indicator);
+//printf("get_payload 1 pid=0x%x unit_start=%d\n", demuxer->pid, demuxer->payload_unit_start_indicator);
 	if(demuxer->payload_unit_start_indicator)
 	{
     	if(demuxer->pid == 0) 
@@ -430,13 +440,24 @@ static int get_payload(mpeg3_demuxer_t *demuxer)
 			printf(" 0x%x bytes elementary data\n", demuxer->raw_size - demuxer->raw_offset);
 		}
 
+/*
+ * if(demuxer->pid == 0x1100) printf("get_payload 1 0x%x\n", demuxer->audio_pid);
+ */
     	if(demuxer->pid == demuxer->audio_pid && 
 			(demuxer->do_audio ||
 			demuxer->read_all))
 		{
 			if(demuxer->do_audio) demuxer->got_audio = 1;
+
+			if(demuxer->dump)
+			{
+				printf("get_payload: offset=%llx 0x%x bytes AC3 pid=0x%x\n", 
+					mpeg3io_tell(demuxer->titles[demuxer->current_title]->fs),
+					demuxer->raw_size - demuxer->raw_offset,
+					demuxer->pid);
+			}
+
 			get_transport_payload(demuxer, 1, 0);
-//printf("get_payload 1 %x\n", demuxer->custom_id);
 		}
     	else 
 		if(demuxer->pid == demuxer->video_pid && 
@@ -447,8 +468,10 @@ static int get_payload(mpeg3_demuxer_t *demuxer)
 			get_transport_payload(demuxer, 0, 1);
 		}
     	else
+		if(demuxer->read_all)
 		{
-			packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
+			get_transport_payload(demuxer, 0, 0);
+//			packet_skip(demuxer, demuxer->raw_size - demuxer->raw_offset);
 		}
 	}
 	return 0;
@@ -479,7 +502,12 @@ demuxer->dump = 0;
 		return 1;
 	}
 
-//printf("read transport 1 %llx %llx\n", title->fs->current_byte, title->fs->total_bytes);
+//fprintf(stderr, "read transport 1 %llx %llx\n", title->fs->current_byte, title->fs->total_bytes);
+
+// Skip BD header
+	if(file->is_bd)
+		mpeg3io_read_int32(title->fs);
+
 // Search for Sync byte */
 	do
 	{
@@ -488,31 +516,36 @@ demuxer->dump = 0;
 
 /* Hit EOF */
 	if(mpeg3io_eof(title->fs) || result) return 1;
-/*
- * printf("read transport 2 bits=%x tell=%llx packet_size=%x\n", 
- * bits, 
- * mpeg3io_tell(title->fs), 
- * file->packet_size);
- */
 
 	if(bits == MPEG3_SYNC_BYTE && !result)
 	{
+// Store byte just read as first byte
 		demuxer->raw_data[0] = MPEG3_SYNC_BYTE;
+
+// Read packet
+		int fragment_size = file->packet_size - 1;
+// Skip BD header
+		if(file->is_bd) 
+		{
+			fragment_size -= 4;
+			demuxer->raw_size -= 4;
+		}
 		result = mpeg3io_read_data(demuxer->raw_data + 1, 
-			file->packet_size - 1, 
+			fragment_size, 
 			title->fs);
 	}
 	else
 	{
-// Assume transport streams are all one program
+// Failed
 		demuxer->program_byte = mpeg3io_tell(title->fs) + 
 			title->start_byte;
 		return 1;
 	}
 
+
+// Sync byte
 	packet_read_char(demuxer);
     bits =  packet_read_int24(demuxer) & 0x00ffffff;
-//printf("read transport 3 tell=%x bits=%x\n", mpeg3io_tell(title->fs), bits);
     demuxer->transport_error_indicator = (bits >> 23) & 0x1;
     demuxer->payload_unit_start_indicator = (bits >> 22) & 0x1;
     demuxer->pid = demuxer->custom_id = (bits >> 8) & 0x00001fff;
@@ -521,20 +554,17 @@ demuxer->dump = 0;
     demuxer->adaptation_field_control = (bits >> 4) & 0x3;
     demuxer->continuity_counter = bits & 0xf;
 
-//printf("read_transport 1 %x\n", demuxer->pid);
 
 // This caused an audio track to not be created.
 	if(demuxer->transport_error_indicator)
 	{
-		fprintf(stderr, 
-			"demuxer->transport_error_indicator at %llx\n", 
-			mpeg3io_tell(title->fs));
+//		fprintf(stderr, "demuxer->transport_error_indicator at %llx\n", 
+//			mpeg3io_tell(title->fs));
 		demuxer->program_byte = mpeg3io_tell(title->fs) + 
 			title->start_byte;
 		return 0;
 	}
 
-//printf("read_transport 5 %x\n", demuxer->pid);
     if (demuxer->pid == 0x1fff)
 	{
 		demuxer->is_padding = 1;  /* padding; just go to next */
@@ -544,7 +574,6 @@ demuxer->dump = 0;
 		demuxer->is_padding = 0;
 	}
 
-//printf("read_transport 6 %x\n", demuxer->pid);
 /* Get pid from table */
 	for(table_entry = 0, result = 0; 
 		table_entry < demuxer->total_pids; 
@@ -557,12 +586,11 @@ demuxer->dump = 0;
 		}
 	}
 
-//printf("read_transport 7 %x\n", demuxer->pid);
 
 
 
 /* Not in pid table */
-	if(!result)
+	if(!result && demuxer->total_pids < MPEG3_PIDMAX)
 	{
 		demuxer->pid_table[table_entry] = demuxer->pid;
 		demuxer->continuity_counters[table_entry] = demuxer->continuity_counter;  /* init */
@@ -572,7 +600,7 @@ demuxer->dump = 0;
 
 	if(demuxer->dump)
 	{
-		printf("0x%llx pid=0x%02x continuity=0x%02x padding=%d adaptation=%d unit_start=%d\n", 
+		printf("offset=0x%llx pid=0x%02x continuity=0x%02x padding=%d adaptation=%d unit_start=%d\n", 
 			demuxer->program_byte,
 			demuxer->pid,
 			demuxer->continuity_counter,
@@ -591,26 +619,6 @@ demuxer->dump = 0;
 	}
 
 
-#if 0
-/* Check counters */
-    if(demuxer->pid != MPEG3_PROGRAM_ASSOCIATION_TABLE && 
-		demuxer->pid != MPEG3_CONDITIONAL_ACCESS_TABLE &&
-        (demuxer->adaptation_field_control == 1 || demuxer->adaptation_field_control == 3))
-	{
-		if(demuxer->continuity_counters[table_entry] != demuxer->continuity_counter)
-		{
-//			fprintf(stderr, "demuxer->continuity_counters[table_entry] != demuxer->continuity_counter\n");
-/* Reset it */
-			demuxer->continuity_counters[table_entry] = demuxer->continuity_counter;
-		}
-		if(++(demuxer->continuity_counters[table_entry]) > 15) demuxer->continuity_counters[table_entry] = 0;
-	}
-#endif
-
-
-
-
-
 
 
     if(demuxer->adaptation_field_control & 0x2)
@@ -774,8 +782,170 @@ static int handle_scrambling(mpeg3_demuxer_t *demuxer,
 	return 0;
 }
 
+static void remove_subtitle(mpeg3_demuxer_t *demuxer, int number)
+{
+	int i, j;
+	mpeg3_delete_subtitle(demuxer->subtitles[number]);
+	for(j = number; j < demuxer->total_subtitles - 1; j++)
+	{
+		demuxer->subtitles[j] = demuxer->subtitles[j + 1];
+	}
+	demuxer->total_subtitles--;
+}
+
+static void remove_subtitle_ptr(mpeg3_demuxer_t *demuxer, mpeg3_subtitle_t *ptr)
+{
+	int i, number = demuxer->total_subtitles;
+	for(i = 0; i < demuxer->total_subtitles; i++)
+	{
+		if(demuxer->subtitles[i] == ptr)
+		{
+			number = i;
+			break;
+		}
+	}
+
+	for(i = number; i < demuxer->total_subtitles - 1; i++)
+	{
+		demuxer->subtitles[i] = demuxer->subtitles[i + 1];
+	}
+	demuxer->total_subtitles--;
+}
+
+/* Create new subtitle object if none with the same id is in the table */
+static mpeg3_subtitle_t* new_subtitle(mpeg3_demuxer_t *demuxer, 
+	int id, 
+	int64_t offset)
+{
+	mpeg3_subtitle_t *subtitle = 0;
+	int i, j;
+
+/* Get subtitle object for the stream and total subtitle objects for the stream */
+	int total = 0;
+	int got_it = 0;
+	for(i = 0; i < demuxer->total_subtitles; i++)
+	{
+		if(demuxer->subtitles[i]->id == id) total++;
+
+		if(demuxer->subtitles[i]->id == id &&
+			!demuxer->subtitles[i]->done &&
+			!got_it)
+		{
+			subtitle = demuxer->subtitles[i];
+			got_it = 1;
+		}
+	}
+
+/* Make new subtitle object */
+	if(!subtitle)
+	{
+		demuxer->total_subtitles++;
+		demuxer->subtitles = realloc(demuxer->subtitles, 
+			sizeof(mpeg3_subtitle_t*) * demuxer->total_subtitles);
+		subtitle = demuxer->subtitles[demuxer->total_subtitles - 1] = 
+			calloc(sizeof(mpeg3_subtitle_t), 1);
+		subtitle->id = id;
+		subtitle->offset = offset;
+	}
+
+/* Delete oldest subtitle */
+	if(total > MPEG3_MAX_SUBTITLES)
+	{
+		for(i = 0; i < demuxer->total_subtitles - 1; i++)
+		{
+			if(demuxer->subtitles[i]->id == id)
+			{
+				remove_subtitle(demuxer, i);
+				break;
+			}
+		}
+	}
+
+	return subtitle;
+}
 
 
+static void handle_subtitle(mpeg3_t *file,
+	mpeg3_demuxer_t *demuxer, 
+	mpeg3_subtitle_t *subtitle,
+	int bytes)
+{
+	mpeg3_title_t *title = demuxer->titles[demuxer->current_title];
+	int i;
+
+/* Allocate memory */
+	if(!subtitle->data)
+	{
+		subtitle->data = malloc(bytes);
+	}
+	else
+	{
+		subtitle->data = realloc(subtitle->data, 
+			subtitle->size + bytes);
+	}
+
+/* Add data to latest subtitle packet. */
+	mpeg3io_read_data(subtitle->data + subtitle->size, 
+		bytes, 
+		title->fs);
+
+/*
+ * static FILE *test = 0;
+ * if(!test) test = fopen("/tmp/debug", "w");
+ * if(stream_id == 0x20) fwrite(subtitle->data + subtitle->size, pes_packet_length, 1, test);
+ * fflush(test);
+ */
+	subtitle->size += bytes;
+
+/* Assume complete subtitle object has been loaded and align it in the buffer */
+	for(i = 0; i < subtitle->size - 1; i++)
+	{
+/* Assume next two bytes are size of the subtitle */
+		unsigned char *ptr = subtitle->data + i;
+		int packet_size = (ptr[0] << 8) | (ptr[1]);
+
+/* i is the first byte of a subtitle packet and the packet is complete */
+		if(i + packet_size <= subtitle->size &&
+			*(ptr + packet_size - 1) == 0xff)
+		{
+
+
+/* Move trailing data to a new object */
+/* This just crashed and we don't think it's supposed to happen. */
+/*
+ * 			if(i + packet_size < subtitle->size)
+ * 			{
+ * 				mpeg3_subtitle_t *subtitle2 = new_subtitle(demuxer, 
+ * 					subtitle->id, 
+ * 					demuxer->program_byte);
+ * 				subtitle2->size = subtitle->size - i - packet_size;
+ * 				subtitle2->data = malloc(subtitle2->size);
+ * 				memcpy(subtitle2->data, subtitle->data + i + packet_size, subtitle2->size);
+ * printf("mpeg3demux handle_subtitle: 2 subtitles in packet subtitle2->size=%d\n", subtitle2->size);
+ * 			}
+ */
+
+/* Move packet to subtitle track */
+			memcpy(subtitle->data, subtitle->data + i, packet_size);
+			subtitle->size = packet_size;
+			subtitle->done = 1;
+
+			remove_subtitle_ptr(demuxer, subtitle);
+			mpeg3_strack_t *strack = mpeg3_create_strack(file, subtitle->id);
+			mpeg3_append_subtitle(strack, subtitle);
+			demuxer->got_subtitle = 1;
+/*
+ * printf("handle_subtitle: id=0x%x pid=%x size=%d offset=0x%llx table_size=%d\n", 
+ * subtitle->id, 
+ * demuxer->pid,
+ * subtitle->size, 
+ * subtitle->offset,
+ * demuxer->total_subtitles);
+ */
+			break;
+		}
+	}
+}
 
 
 
@@ -871,7 +1041,7 @@ static int handle_pcm(mpeg3_demuxer_t *demuxer, int bytes)
 
 
 /* Program packet reading core */
-static int get_ps_pes_packet(mpeg3_demuxer_t *demuxer, unsigned int header)
+static int get_program_pes_packet(mpeg3_demuxer_t *demuxer, unsigned int header)
 {
 	unsigned int pts = 0, dts = 0;
 	int pes_packet_length;
@@ -969,7 +1139,7 @@ static int get_ps_pes_packet(mpeg3_demuxer_t *demuxer, unsigned int header)
         		pes_header_bytes += 10;
     		}
 
-//printf("get_ps_pes_packet do_audio=%d do_video=%d pts=%x dts=%x\n", 
+//printf("get_program_pes_packet do_audio=%d do_video=%d pts=%x dts=%x\n", 
 //	demuxer->do_audio, demuxer->do_video, pts, dts);
 
 /* Skip unknown */
@@ -1110,10 +1280,29 @@ static int get_ps_pes_packet(mpeg3_demuxer_t *demuxer, unsigned int header)
     			mpeg3io_seek_relative(title->fs, pes_packet_length);
     		}
     	}
-    	else 
+    	else
 		if((demuxer->stream_id == 0xbd || demuxer->stream_id == 0xbf) && 
 			mpeg3io_next_char(title->fs) != 0xff &&
-			((mpeg3io_next_char(title->fs) & 0xf0) != 0x20))
+			((mpeg3io_next_char(title->fs) & 0xf0) == 0x20))
+		{
+/* DVD subtitle data */
+			mpeg3_subtitle_t *subtitle = 0;
+			int stream_id = demuxer->stream_id = mpeg3io_read_char(title->fs);
+			subtitle = new_subtitle(demuxer, 
+				stream_id, 
+				demuxer->program_byte);
+
+/* Get data length */
+			pes_packet_length -= mpeg3io_tell(title->fs) - pes_packet_start;
+
+			handle_subtitle(file, demuxer, subtitle, pes_packet_length);
+//printf("mpeg3_demux id=0x%02x size=%d total_size=%d\n", stream_id, pes_packet_length, subtitle->size);
+
+		}
+		else
+		if((demuxer->stream_id == 0xbd || demuxer->stream_id == 0xbf) && 
+			mpeg3io_next_char(title->fs) != 0xff &&
+			((mpeg3io_next_char(title->fs) & 0xf0) == 0x80))
 		{
 			int format;
 /* DVD audio data */
@@ -1156,7 +1345,7 @@ static int get_ps_pes_packet(mpeg3_demuxer_t *demuxer, unsigned int header)
 				decryption_offset = mpeg3io_tell(title->fs) - demuxer->last_packet_start;
 
 				if(format == AUDIO_PCM) do_pcm = 1;
-//printf("get_ps_pes_packet 5 %x\n", decryption_offset);
+//printf("get_program_pes_packet 5 %x\n", decryption_offset);
 
 				if(demuxer->dump)
 				{
@@ -1173,7 +1362,7 @@ static int get_ps_pes_packet(mpeg3_demuxer_t *demuxer, unsigned int header)
 				pes_packet_length -= mpeg3io_tell(title->fs) - pes_packet_start;
     			mpeg3io_seek_relative(title->fs, pes_packet_length);
       		}
-//printf("get_ps_pes_packet 6 %d\n", demuxer->astream_table[0x20]);
+//printf("get_program_pes_packet 6 %d\n", demuxer->astream_table[0x20]);
     	}
     	else 
 		if(demuxer->stream_id == 0xbc || 1)
@@ -1221,6 +1410,7 @@ int mpeg3demux_read_program(mpeg3_demuxer_t *demuxer)
 	demuxer->got_video = 0;
 	demuxer->stream_id = 0;
 	demuxer->custom_id = -1;
+	demuxer->got_subtitle = 0;
 
 	if(mpeg3io_eof(title->fs)) return 1;
 
@@ -1273,7 +1463,7 @@ int mpeg3demux_read_program(mpeg3_demuxer_t *demuxer)
 		{
 
 
-			result = get_ps_pes_packet(demuxer, header);
+			result = get_program_pes_packet(demuxer, header);
 
 
 
@@ -1424,8 +1614,8 @@ int mpeg3_seek_phys(mpeg3_demuxer_t *demuxer)
 
 	if(!demuxer->titles) return 1;
 
-//printf("%d %d\n", demuxer->current_title, demuxer->title_cell);
 	mpeg3_title_t *title = demuxer->titles[demuxer->current_title];
+//printf("mpeg3_seek_phys: %p %d %d\n", title->cell_table, demuxer->current_title, demuxer->title_cell);
 
 	if(!title->cell_table) return 1;
 
@@ -1445,6 +1635,7 @@ int mpeg3_seek_phys(mpeg3_demuxer_t *demuxer)
 	else
 	{
 // End of stream
+//printf("mpeg3_seek_phys %lld %lld\n", demuxer->program_byte, demuxer->stream_end);
 		if(demuxer->stream_end > 0 &&
 			demuxer->program_byte >= demuxer->stream_end) return 1;
 
@@ -1529,7 +1720,6 @@ int mpeg3_read_next_packet(mpeg3_demuxer_t *demuxer)
 	mpeg3_title_t *title = demuxer->titles[demuxer->current_title];
 
 
-//printf("mpeg3_read_next_packet 1 %lld\n", demuxer->program_byte);
 /* Reset output buffer */
 	demuxer->data_size = 0;
 	demuxer->data_position = 0;
@@ -1580,6 +1770,7 @@ int mpeg3_read_next_packet(mpeg3_demuxer_t *demuxer)
 		do
 		{
 			title = demuxer->titles[demuxer->current_title];
+//printf("mpeg3_read_next_packet 10 %llx\n", mpeg3io_tell(title->fs));
 
 			if(!result)
 			{
@@ -1620,9 +1811,16 @@ int mpeg3_read_next_packet(mpeg3_demuxer_t *demuxer)
 						file->packet_size, title->fs);
 					demuxer->data_size = file->packet_size;
 					demuxer->program_byte += file->packet_size;
+//printf("mpeg3_read_next_packet 1 %llx\n", mpeg3demux_tell_byte(demuxer));
 					result |= mpeg3_seek_phys(demuxer);
+//printf("mpeg3_read_next_packet 2 %llx\n", mpeg3demux_tell_byte(demuxer));
 				}
 			}
+/*
+ * printf("mpeg3_read_next_packet 100 result=%d data_size=0x%x\n", 
+ * result,
+ * demuxer->data_size);
+ */
 		}while(!result && 
 			demuxer->data_size == 0 && 
 			(demuxer->do_audio || demuxer->do_video));
@@ -1673,7 +1871,6 @@ int mpeg3_read_prev_packet(mpeg3_demuxer_t *demuxer)
 
  
 
-//printf("mpeg3_read_prev_packet 1 %d %d %llx\n", result, demuxer->current_title, mpeg3io_tell(title->fs));
 /* Switch to reverse direction */
 	if(!demuxer->reverse)
 	{
@@ -1708,8 +1905,10 @@ int mpeg3_read_prev_packet(mpeg3_demuxer_t *demuxer)
 /* Transport stream or elementary stream case */
 		if(file->packet_size > 0)
 		{
+printf("mpeg3_read_prev_packet 1 result=%d title=%d tell=%llx program_byte=%llx\n", result, demuxer->current_title, mpeg3io_tell(title->fs), demuxer->program_byte);
 			demuxer->program_byte -= file->packet_size;
 			result = mpeg3_seek_phys(demuxer);
+printf("mpeg3_read_prev_packet 100 result=%d title=%d tell=%llx program_byte=%llx\n", result, demuxer->current_title, mpeg3io_tell(title->fs), demuxer->program_byte);
 		}
 		else
 		{
@@ -1778,6 +1977,7 @@ int mpeg3demux_read_data(mpeg3_demuxer_t *demuxer,
 {
 	int result = 0;
 	demuxer->error_flag = 0;
+//printf("mpeg3demux_read_data 1\n");
 
 	if(demuxer->data_position >= 0)
 	{
@@ -1796,7 +1996,15 @@ int mpeg3demux_read_data(mpeg3_demuxer_t *demuxer,
 			{
 				result = mpeg3_read_next_packet(demuxer);
 			}
+/*
+ * printf("mpeg3demux_read_data 10 offset=%llx pid=0x%x bytes=0x%x i=0x%x\n", 
+ * mpeg3demux_tell_byte(demuxer),
+ * demuxer->pid, 
+ * demuxer->data_size,
+ * i);
+ */
 		}
+//printf("mpeg3demux_read_data 10\n");
 	}
 	else
 	{
@@ -1894,11 +2102,13 @@ int mpeg3demux_copy_titles(mpeg3_demuxer_t *dst, mpeg3_demuxer_t *src)
 
 	dst->total_titles = src->total_titles;
 	dst->total_programs = src->total_programs;
+
 	for(i = 0; i < MPEG3_MAX_STREAMS; i++)
 	{
 		dst->astream_table[i] = src->astream_table[i];
 		dst->vstream_table[i] = src->vstream_table[i];
 	}
+
 	for(i = 0; i < src->total_titles; i++)
 	{
 		src_title = src->titles[i];
@@ -1969,6 +2179,11 @@ int mpeg3_delete_demuxer(mpeg3_demuxer_t *demuxer)
 	if(demuxer->raw_data) free(demuxer->raw_data);
 	if(demuxer->audio_buffer) free(demuxer->audio_buffer);
 	if(demuxer->video_buffer) free(demuxer->video_buffer);
+	for(i = 0; i < demuxer->total_subtitles; i++)
+	{
+		mpeg3_delete_subtitle(demuxer->subtitles[i]);
+	}
+	if(demuxer->subtitles) free(demuxer->subtitles);
 	free(demuxer);
 	return 0;
 }
@@ -1985,6 +2200,10 @@ int mpeg3demux_eof(mpeg3_demuxer_t *demuxer)
 				demuxer->current_title >= demuxer->total_titles - 1)
 				return 1;
 		}
+
+// Same operation performed in mpeg3_seek_phys
+		if(demuxer->stream_end > 0 &&
+			demuxer->program_byte >= demuxer->stream_end) return 1;
 	}
 	else
 	{
@@ -2046,6 +2265,19 @@ int mpeg3demux_seek_byte(mpeg3_demuxer_t *demuxer, int64_t byte)
 }
 
 
+int mpeg3_finished_subtitles(mpeg3_demuxer_t *demuxer, int id)
+{
+	int i;
+	int total = 0;
+	for(i = 0; i < demuxer->total_subtitles; i++)
+	{
+		if(demuxer->subtitles[i]->done &&
+			(id < 0 ||
+			demuxer->subtitles[i]->id == id)) total++;
+	}
+	return total;
+}
+
 
 
 
