@@ -5,6 +5,48 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+#include <sys/stat.h>
+
+
+
+#define PUT_INT32(x) \
+{ \
+	uint32_t temp = x; \
+	if(MPEG3_LITTLE_ENDIAN) \
+	{ \
+		fputc(((unsigned char*)&temp)[3], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[2], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[1], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[0], file->toc_fd); \
+	} \
+	else \
+	{ \
+		fwrite(&temp, 1, 4, file->toc_fd); \
+	} \
+}
+
+
+
+
+#define PUT_INT64(x) \
+{ \
+	uint64_t temp = x; \
+	if(MPEG3_LITTLE_ENDIAN) \
+	{ \
+		fputc(((unsigned char*)&temp)[7], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[6], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[5], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[4], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[3], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[2], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[1], file->toc_fd); \
+		fputc(((unsigned char*)&temp)[0], file->toc_fd); \
+	} \
+	else \
+	{ \
+		fwrite(&temp, 1, 8, file->toc_fd); \
+	} \
+}
 
 
 static uint32_t read_int32(unsigned char *buffer, int *position)
@@ -29,6 +71,7 @@ static uint32_t read_int32(unsigned char *buffer, int *position)
 	return temp;
 }
 
+
 static uint64_t read_int64(unsigned char *buffer, int *position)
 {
 	uint64_t temp;
@@ -68,9 +111,32 @@ static int read_data(unsigned char *buffer,
 	*position += bytes;
 }
 
-int mpeg3_read_toc(mpeg3_t *file, int *atracks_return, int *vtracks_return)
+// Concatenate title and toc directory if title is not absolute and
+// toc path has a directory section.
+static void complete_path(char *result, char *toc_path, char *path)
+{
+	strcpy(result, path);
+
+	if(path[0] != '/')
+	{
+// Get end of toc directory
+		char *ptr = strrchr(toc_path, '/');
+
+		if(ptr)
+		{
+// Stack filename on toc path
+			strcpy(result, toc_path);
+			strcpy(&result[ptr - toc_path + 1], path);
+		}
+	}
+}
+
+int mpeg3_read_toc(mpeg3_t *file, 
+	int *atracks_return, 
+	int *vtracks_return)
 {
 	unsigned char *buffer;
+	int buffer_size;
 	int file_type;
 	int position = 4;
 	int stream_type;
@@ -79,283 +145,349 @@ int mpeg3_read_toc(mpeg3_t *file, int *atracks_return, int *vtracks_return)
 	int vfs_len = strlen(RENDERFARM_FS_PREFIX);
 	int toc_version;
 	int64_t current_byte = 0;
+	char *ext;
+const int debug = 0;
 
-//printf("read_toc 1\n");
 // Fix title paths for Cinelerra VFS
 	if(!strncmp(file->fs->path, RENDERFARM_FS_PREFIX, vfs_len))
 		is_vfs = 1;
 
-	buffer = malloc(mpeg3io_total_bytes(file->fs));
+	buffer_size = mpeg3io_total_bytes(file->fs);
+	buffer = malloc(buffer_size);
 	mpeg3io_seek(file->fs, 0);
-	mpeg3io_read_data(buffer, mpeg3io_total_bytes(file->fs), file->fs);
+	mpeg3io_read_data(buffer, buffer_size, file->fs);
 
 // Test version
-	if((toc_version = buffer[position++]) != MPEG3_TOC_VERSION)
+	if((toc_version = read_int32(buffer, &position)) != MPEG3_TOC_VERSION)
 	{
+		free(buffer);
 		fprintf(stderr,
-			"read_toc: invalid TOC version %x\n", 
+			"mpeg3_read_toc: invalid TOC version %x\n", 
 			toc_version);
-		return 1;
+		return MPEG3_INVALID_TOC_VERSION;
 	}
 
-//printf("read_toc %lld\n", mpeg3io_total_bytes(file->fs));
 
 // File type
-	file_type = buffer[position++];
-	switch(file_type)
+	int done = 0;
+	while(!done && position <= buffer_size - 4)
 	{
-		case FILE_TYPE_PROGRAM:
-			file->is_program_stream = 1;
-			break;
-		case FILE_TYPE_TRANSPORT:
-			file->is_transport_stream = 1;
-			break;
-		case FILE_TYPE_AUDIO:
-			file->is_audio_stream = 1;
-			break;
-		case FILE_TYPE_VIDEO:
-			file->is_video_stream = 1;
-			break;
-	}
-
-//printf("read_toc 10\n");
-
-// Stream ID's
-	while((stream_type = buffer[position]) != TITLE_PATH)
-	{
-		int offset;
-		int stream_id;
-
-//printf("read_toc %d %x\n", position, buffer[position]);
-		position++;
-		offset = read_int32(buffer, &position);
-		stream_id = read_int32(buffer, &position);
-
-		if(stream_type == STREAM_AUDIO)
+// Get section type
+		int section_type = read_int32(buffer, &position);
+//printf("section_type=%d position=%x\n", section_type, position);
+		switch(section_type)
 		{
-			file->demuxer->astream_table[offset] = stream_id;
-		}
+			case FILE_TYPE_PROGRAM:
+				file->is_program_stream = 1;
+				break;
+			case FILE_TYPE_TRANSPORT:
+				file->is_transport_stream = 1;
+				break;
+			case FILE_TYPE_AUDIO:
+				file->is_audio_stream = 1;
+				break;
+			case FILE_TYPE_VIDEO:
+				file->is_video_stream = 1;
+				break;
 
-		if(stream_type == STREAM_VIDEO)
-		{
-			file->demuxer->vstream_table[offset] = stream_id;
-		}
-	}
+			case FILE_INFO:
+			{
+				char string[MPEG3_STRLEN];
+				char string2[MPEG3_STRLEN];
+				memcpy(string, &buffer[position], MPEG3_STRLEN);
+				complete_path(string2, file->fs->path, string);
+
+				position += MPEG3_STRLEN;
+				file->source_date = read_int64(buffer, &position);
+				int64_t current_date = mpeg3_calculate_source_date(string2);
+/*
+ * printf("mpeg3_read_toc file=%s source_date=%lld current_date=%lld\n", 
+ * string2, 
+ * file->source_date,
+ * current_date);
+ */
+				if(current_date != file->source_date)
+				{
+					fprintf(stderr, "read_toc: date mismatch\n");
+					free(buffer);
+					return MPEG3_TOC_DATE_MISMATCH;
+				}
+				break;
+			}
+
+			case STREAM_AUDIO:
+			{
+				int number;
+				int stream_id;
+				number = read_int32(buffer, &position);
+				stream_id = read_int32(buffer, &position);
+				file->demuxer->astream_table[number] = stream_id;
+				break;
+			}
+
+			case STREAM_VIDEO:
+			{
+				int number;
+				int stream_id;
+
+				number = read_int32(buffer, &position);
+				stream_id = read_int32(buffer, &position);
+				file->demuxer->vstream_table[number] = stream_id;
+				break;
+			}
 
 
+			case ATRACK_COUNT:
+				*atracks_return = read_int32(buffer, &position);
+				file->channel_counts = calloc(sizeof(int), *atracks_return);
+				file->sample_offsets = calloc(sizeof(int64_t*), *atracks_return);
+				file->total_sample_offsets = calloc(sizeof(int), *atracks_return);
+				file->audio_eof = calloc(sizeof(int64_t), *atracks_return);
+				file->total_samples = calloc(sizeof(int64_t), *atracks_return);
+				file->indexes = calloc(sizeof(mpeg3_index_t*), *atracks_return);
+				file->total_indexes = *atracks_return;
+				for(i = 0; i < *atracks_return; i++)
+				{
+					file->audio_eof[i] = read_int64(buffer, &position);
+					file->channel_counts[i] = read_int32(buffer, &position);
+					file->total_sample_offsets[i] = read_int32(buffer, &position);
+					file->total_samples[i] = read_int64(buffer, &position);
 
-//printf("read_toc 10\n");
+					if(file->total_samples[i] < 1) file->total_samples[i] = 1;
+					file->sample_offsets[i] = malloc(file->total_sample_offsets[i] * sizeof(int64_t));
+					for(j = 0; j < file->total_sample_offsets[i]; j++)
+					{
+						file->sample_offsets[i][j] = read_int64(buffer, &position);
+					}
 
+					mpeg3_index_t *index = file->indexes[i] = mpeg3_new_index();
+					index->index_size = read_int32(buffer, &position);
+					index->index_zoom = read_int32(buffer, &position);
+//printf("mpeg3_read_toc %d %d %d\n", i, index->index_size, index->index_zoom);
+					int channels = index->index_channels = file->channel_counts[i];
+					if(channels)
+					{
+						index->index_data = calloc(sizeof(float*), channels);
+						for(j = 0; j < channels; j++)
+						{
+							index->index_data[j] = calloc(sizeof(float), 
+								index->index_size * 2);
+							read_data(buffer,
+								&position,
+								(unsigned char*)index->index_data[j], 
+								sizeof(float) * index->index_size * 2);
+						}
+					}
+				}
+				break;
 
-// Titles
-	while(buffer[position] == TITLE_PATH)
-	{
-		char string[MPEG3_STRLEN];
-		int string_len = 0;
-		mpeg3_title_t *title;
-		FILE *test_fd;
+			case VTRACK_COUNT:
+				*vtracks_return = read_int32(buffer, &position);
+				file->frame_offsets = calloc(sizeof(int64_t*), *vtracks_return);
+				file->total_frame_offsets = calloc(sizeof(int), *vtracks_return);
+				file->keyframe_numbers = calloc(sizeof(int64_t*), *vtracks_return);
+				file->total_keyframe_numbers = calloc(sizeof(int), *vtracks_return);
+				file->video_eof = calloc(sizeof(int64_t), *vtracks_return);
+				for(i = 0; i < *vtracks_return; i++)
+				{
+					file->video_eof[i] = read_int64(buffer, &position);
+					file->total_frame_offsets[i] = read_int32(buffer, &position);
+					file->frame_offsets[i] = malloc(file->total_frame_offsets[i] * sizeof(int64_t));
+if(debug) printf("mpeg3_read_toc 62 %d %d %lld\n", 
+file->total_frame_offsets[i], position, buffer_size);
+					for(j = 0; j < file->total_frame_offsets[i]; j++)
+					{
+						file->frame_offsets[i][j] = read_int64(buffer, &position);
+//printf("frame %llx\n", file->frame_offsets[i][j]);
+					}
+
+if(debug) printf("mpeg3_read_toc 64\n");
+
+					file->total_keyframe_numbers[i] = read_int32(buffer, &position);
+					file->keyframe_numbers[i] = malloc(file->total_keyframe_numbers[i] * sizeof(int64_t));
+					for(j = 0; j < file->total_keyframe_numbers[i]; j++)
+					{
+						file->keyframe_numbers[i][j] = read_int64(buffer, &position);
+					}
+				}
+				break;
+
+			case STRACK_COUNT:
+			{
+				file->total_sstreams = read_int32(buffer, &position);
+				for(i = 0; i < file->total_sstreams; i++)
+				{
+					int id = read_int32(buffer, &position);
+					mpeg3_strack_t *strack = file->strack[i] = mpeg3_new_strack(id);
+					strack->total_offsets = read_int32(buffer, &position);
+					strack->offsets = malloc(sizeof(int64_t) * strack->total_offsets);
+					strack->allocated_offsets = strack->total_offsets;
+					for(j = 0; j < strack->total_offsets; j++)
+					{
+						strack->offsets[j] = read_int64(buffer, &position);
+					}
+				}
+				break;
+			}
+
+			case TITLE_PATH:
+			{
+				char string[MPEG3_STRLEN];
+				int string_len = 0;
+				mpeg3_title_t *title;
+				FILE *test_fd;
+if(debug) printf("mpeg3_read_toc 11\n");
 
 // Construct title path from VFS prefix and path.
-		position++;
-		if(is_vfs)
-		{
-			strcpy(string, RENDERFARM_FS_PREFIX);
-			string_len = vfs_len;
-		}
-		while(buffer[position] != 0) string[string_len++] = buffer[position++];
-		string[string_len++] = 0;
-		position++;
+				if(is_vfs)
+				{
+					strcpy(string, RENDERFARM_FS_PREFIX);
+					string_len = vfs_len;
+				}
+				memcpy(string, buffer + position, MPEG3_STRLEN);
+				position += MPEG3_STRLEN;
+
+// Detect Blu-Ray
+if(debug) printf("mpeg3_read_toc 11 position=%x\n", position);
+				ext = strrchr(string, '.');
+				if(ext && !strncasecmp(ext, ".m2ts", 5))
+					file->is_bd = 1;
+if(debug) printf("mpeg3_read_toc 12\n");
 
 // Test title availability
-		test_fd = fopen(string, "r");
-		if(test_fd) 
-		{
-			fclose(test_fd);
-		}
-		else
-		{
-// Try concatenating title and toc directory if title is not absolute and
-// toc path has a directory section.
-			if((!is_vfs && string[0] != '/') ||
-				(is_vfs && string[vfs_len] != '/'))
-			{
-// Get toc filename without path
-				char *ptr = strrchr(file->fs->path, '/');
-				if(ptr)
+				test_fd = fopen(string, "r");
+if(debug) printf("mpeg3_read_toc 20\n");
+				if(test_fd) 
 				{
-					char string2[MPEG3_STRLEN];
+					fclose(test_fd);
+				}
+				else
+				{
+// Concatenate title and toc directory if title is not absolute and
+// toc path has a directory section.
+					if((!is_vfs && string[0] != '/') ||
+						(is_vfs && string[vfs_len] != '/'))
+					{
+// Get toc filename without path
+						char *ptr = strrchr(file->fs->path, '/');
+						if(ptr)
+						{
+							char string2[MPEG3_STRLEN];
 
 // Stack filename on toc path
-					strcpy(string2, file->fs->path);
-					if(!is_vfs)
-						strcpy(&string2[ptr - file->fs->path + 1], string);
-					else
-						strcpy(&string2[ptr - file->fs->path + 1], string + vfs_len);
+							strcpy(string2, file->fs->path);
+							if(!is_vfs)
+								strcpy(&string2[ptr - file->fs->path + 1], string);
+							else
+								strcpy(&string2[ptr - file->fs->path + 1], string + vfs_len);
 
-					test_fd = fopen(string2, "r");
-					if(test_fd)
-					{
-						fclose(test_fd);
-						strcpy(string, string2);
+							test_fd = fopen(string2, "r");
+							if(test_fd)
+							{
+								fclose(test_fd);
+								strcpy(string, string2);
+							}
+							else
+							{
+								fprintf(stderr, 
+									"read_toc: failed to open %s or %s\n",
+									string,
+									string2);
+								return 1;
+							}
+						}
+						else
+						{
+							fprintf(stderr,
+								"read_toc: failed to open %s\n", 
+								string);
+							return 1;
+						}
 					}
 					else
 					{
 						fprintf(stderr, 
-							"read_toc: failed to open %s or %s\n",
-							string,
-							string2);
+							"read_toc: failed to open %s\n", 
+							string);
 						return 1;
 					}
 				}
-				else
-				{
-					fprintf(stderr,
-						"read_toc: failed to open %s\n", 
-						string);
-					return 1;
-				}
-			}
-			else
-			{
-				fprintf(stderr, 
-					"read_toc: failed to open %s\n", 
-					string);
-				return 1;
-			}
-		}
 
-		title = 
-			file->demuxer->titles[file->demuxer->total_titles++] = 
-			mpeg3_new_title(file, string);
+if(debug) printf("mpeg3_read_toc 30\n");
+				title = 
+					file->demuxer->titles[file->demuxer->total_titles++] = 
+					mpeg3_new_title(file, string);
 
-		title->total_bytes = read_int64(buffer, &position);
-		title->start_byte = current_byte;
-		title->end_byte = title->start_byte + title->total_bytes;
-		current_byte = title->end_byte;
+				title->total_bytes = read_int64(buffer, &position);
+				title->start_byte = current_byte;
+				title->end_byte = title->start_byte + title->total_bytes;
+				current_byte = title->end_byte;
 
 // Cells
-		title->cell_table_size = 
-			title->cell_table_allocation = 
-			read_int32(buffer, &position);
-		title->cell_table = calloc(title->cell_table_size, sizeof(mpeg3_cell_t));
-		for(i = 0; i < title->cell_table_size; i++)
-		{
-			mpeg3_cell_t *cell = &title->cell_table[i];
-			cell->title_start = read_int64(buffer, &position);
-			cell->title_end = read_int64(buffer, &position);
-			cell->program_start = read_int64(buffer, &position);
-			cell->program_end = read_int64(buffer, &position);
-			cell->program = read_int32(buffer, &position);
-		}
-	}
-//printf("read_toc 10\n");
+				title->cell_table_size = 
+					title->cell_table_allocation = 
+					read_int32(buffer, &position);
+//printf("mpeg3_read_toc 40 %llx %d\n", title->total_bytes, title->cell_table_size);
+				title->cell_table = calloc(title->cell_table_size, sizeof(mpeg3_cell_t));
+				for(i = 0; i < title->cell_table_size; i++)
+				{
+					mpeg3_cell_t *cell = &title->cell_table[i];
+					cell->title_start = read_int64(buffer, &position);
+					cell->title_end = read_int64(buffer, &position);
+					cell->program_start = read_int64(buffer, &position);
+					cell->program_end = read_int64(buffer, &position);
+					cell->program = read_int32(buffer, &position);
+/*
+ * printf("mpeg3_read_toc 50 %llx-%llx %llx-%llx %d\n", 
+ * cell->title_start, 
+ * cell->title_end,
+ * cell->program_start,
+ * cell->program_end,
+ * cell->program);
+ */
+				}
 
-
-
-// Audio streams
-// Skip ATRACK_COUNT
-	position++;
-	*atracks_return = read_int32(buffer, &position);
-//printf("read_toc 1 %d\n", *atracks_return);
-
-// Skip VTRACK_COUNT
-	position++;
-	*vtracks_return = read_int32(buffer, &position);
-//printf("read_toc 2 %d\n", *vtracks_return);
-
-
-	if(*atracks_return)
-	{
-		file->channel_counts = calloc(sizeof(int), *atracks_return);
-		file->sample_offsets = calloc(sizeof(int64_t*), *atracks_return);
-		file->total_sample_offsets = calloc(sizeof(int*), *atracks_return);
-		file->audio_eof = calloc(sizeof(int64_t), *atracks_return);
-
-		for(i = 0; i < *atracks_return; i++)
-		{
-			file->audio_eof[i] = read_int64(buffer, &position);
-			file->channel_counts[i] = read_int32(buffer, &position);
-			file->total_sample_offsets[i] = read_int32(buffer, &position);
-			file->sample_offsets[i] = malloc(file->total_sample_offsets[i] * sizeof(int64_t));
-			for(j = 0; j < file->total_sample_offsets[i]; j++)
-			{
-				file->sample_offsets[i][j] = read_int64(buffer, &position);
-			}
-//printf("samples %d %x %llx\n", i, file->sample_offsets[i][0]);
-		}
-	}
-//printf("read_toc 10\n");
-
-	if(*vtracks_return)
-	{
-		file->frame_offsets = calloc(sizeof(int64_t*), *vtracks_return);
-		file->total_frame_offsets = calloc(sizeof(int*), *vtracks_return);
-		file->keyframe_numbers = calloc(sizeof(int64_t*), *vtracks_return);
-		file->total_keyframe_numbers = calloc(sizeof(int*), *vtracks_return);
-		file->video_eof = calloc(sizeof(int64_t), *vtracks_return);
-
-		for(i = 0; i < *vtracks_return; i++)
-		{
-			file->video_eof[i] = read_int64(buffer, &position);
-			file->total_frame_offsets[i] = read_int32(buffer, &position);
-			file->frame_offsets[i] = malloc(file->total_frame_offsets[i] * sizeof(int64_t));
-			for(j = 0; j < file->total_frame_offsets[i]; j++)
-			{
-				file->frame_offsets[i][j] = read_int64(buffer, &position);
-//printf("frame %llx\n", file->frame_offsets[i][j]);
+				break;
 			}
 
-
-			file->total_keyframe_numbers[i] = read_int32(buffer, &position);
-			file->keyframe_numbers[i] = malloc(file->total_keyframe_numbers[i] * sizeof(int64_t));
-			for(j = 0; j < file->total_keyframe_numbers[i]; j++)
-			{
-				file->keyframe_numbers[i][j] = read_int64(buffer, &position);
-			}
+			case IFO_PALETTE:
+				for(i = 0; i < 16; i++)
+				{
+					file->palette[i * 4 + 0] = (unsigned char)buffer[position++];
+					file->palette[i * 4 + 1] = (unsigned char)buffer[position++];
+					file->palette[i * 4 + 2] = (unsigned char)buffer[position++];
+					file->palette[i * 4 + 3] = (unsigned char)buffer[position++];
+/*
+ * printf("color %02d: 0x%02x 0x%02x 0x%02x 0x%02x\n", 
+ * i,
+ * file->palette[i * 4 + 0], 
+ * file->palette[i * 4 + 1], 
+ * file->palette[i * 4 + 2], 
+ * file->palette[i * 4 + 3]);
+ */
+				}
+				file->have_palette = 1;
+				break;
 		}
 	}
 
 
 
-// Indexes.  ATracks aren't created yet so we need to put the indexes somewhere
-// else.
-	file->indexes = calloc(sizeof(mpeg3_index_t*), *atracks_return);
-	file->total_indexes = *atracks_return;
-	for(i = 0; i < *atracks_return; i++)
-	{
-		mpeg3_index_t *index = file->indexes[i] = mpeg3_new_index();
-		
-		index->index_size = read_int32(buffer, &position);
-		index->index_zoom = read_int32(buffer, &position);
-//printf("mpeg3_read_toc %d %d %d\n", i, index->index_size, index->index_zoom);
-		int channels = index->index_channels = file->channel_counts[i];
-		if(channels)
-		{
-			index->index_data = calloc(sizeof(float*), channels);
-			for(j = 0; j < channels; j++)
-			{
-				index->index_data[j] = calloc(sizeof(float), 
-					index->index_size * 2);
-				read_data(buffer,
-					&position,
-					(unsigned char*)index->index_data[j], 
-					sizeof(float) * index->index_size * 2);
-			}
-		}
-	}
+
+
+
 
 
 
 
 
 	free(buffer);
-//printf("read_toc 10\n");
+if(debug) printf("mpeg3_read_toc 90\n");
 
 
 
-//printf("read_toc 1\n");
 	mpeg3demux_open_title(file->demuxer, 0);
-//printf("read_toc 10\n");
+if(debug) printf("mpeg3_read_toc 100\n");
 
 	return 0;
 }
@@ -366,6 +498,7 @@ mpeg3_t* mpeg3_start_toc(char *path, char *toc_path, int64_t *total_bytes)
 	*total_bytes = 0;
 	mpeg3_t *file = mpeg3_new(path);
 
+
 	file->toc_fd = fopen(toc_path, "w");
 	if(!file->toc_fd)
 	{
@@ -377,7 +510,7 @@ mpeg3_t* mpeg3_start_toc(char *path, char *toc_path, int64_t *total_bytes)
 	}
 	
 	
-
+	file->source_date = mpeg3_calculate_source_date(path);
 	file->seekable = 0;
 
 /* Authenticate encryption before reading a single byte */
@@ -389,7 +522,11 @@ mpeg3_t* mpeg3_start_toc(char *path, char *toc_path, int64_t *total_bytes)
 
 // Determine file type
 	int toc_atracks = 0, toc_vtracks = 0;
-	if(mpeg3_get_file_type(file, 0, 0, 0)) return 0;
+	if(mpeg3_get_file_type(file, 0, 0, 0))
+	{
+		mpeg3_delete(file);
+		return 0;
+	}
 
 
 
@@ -416,6 +553,7 @@ mpeg3_t* mpeg3_start_toc(char *path, char *toc_path, int64_t *total_bytes)
 	file->demuxer->read_all = 1;
 	*total_bytes = mpeg3demux_movie_size(file->demuxer);
 
+//*total_bytes = 500000000;
 	return file;
 }
 
@@ -594,7 +732,7 @@ int mpeg3_update_index(mpeg3_t *file,
 
 
 
-int mpeg3_toc_audio(mpeg3_t *file, 
+static int handle_audio(mpeg3_t *file, 
 	int track_number)
 {
 	int i, j, k;
@@ -604,10 +742,34 @@ int mpeg3_toc_audio(mpeg3_t *file,
 	atrack->audio_eof = mpeg3demux_tell_byte(file->demuxer);
 
 // Append demuxed data to track buffer
-	mpeg3demux_append_data(atrack->demuxer,
-		file->demuxer->audio_buffer,
-		file->demuxer->audio_size);
+	if(file->demuxer->audio_size)
+		mpeg3demux_append_data(atrack->demuxer,
+			file->demuxer->audio_buffer,
+			file->demuxer->audio_size);
+	else
+	if(file->demuxer->data_size)
+		mpeg3demux_append_data(atrack->demuxer,
+			file->demuxer->data_buffer,
+			file->demuxer->data_size);
 
+/*
+ * if(file->demuxer->pid == 0x1100) printf("handle_audio %p %d %d\n", 
+ * file->demuxer->pid,
+ * file->demuxer->audio_size, 
+ * file->demuxer->data_size);
+ */
+
+/*
+ * if(atrack->pid == 0x1100)
+ * {
+ * static FILE *test = 0;
+ * if(!test) test = fopen("/hmov/test.ac3", "w");
+ * fwrite(file->demuxer->audio_buffer,
+ * file->demuxer->audio_size,
+ * 1,
+ * test);
+ * }
+ */
 
 // Decode samples
 	mpeg3audio_decode_audio(atrack->audio, 
@@ -623,25 +785,26 @@ int mpeg3_toc_audio(mpeg3_t *file,
 	return 0;
 }
 
-int mpeg3_toc_video(mpeg3_t *file, 
+
+static int handle_video(mpeg3_t *file, 
 	mpeg3_vtrack_t *vtrack)
 {
 	mpeg3video_t *video = vtrack->video;
 
-/*
- * static FILE *test = 0;
- * if(!test) test = fopen("test.m2v", "w");
- * fwrite(file->demuxer->data_buffer, 1, file->demuxer->data_size, test);
- * static int counter = 0;
- * printf("%x %d\n", vtrack->pid, counter++);
- */
+
 // Assume last packet of stream
 	vtrack->video_eof = mpeg3demux_tell_byte(file->demuxer);
 
 // Append demuxed data to track buffer
-	mpeg3demux_append_data(vtrack->demuxer,
-		file->demuxer->video_buffer,
-		file->demuxer->video_size);
+	if(file->demuxer->video_size)
+		mpeg3demux_append_data(vtrack->demuxer,
+			file->demuxer->video_buffer,
+			file->demuxer->video_size);
+	else
+	if(file->demuxer->data_size)
+		mpeg3demux_append_data(vtrack->demuxer,
+			file->demuxer->data_buffer,
+			file->demuxer->data_size);
 
 
 	if(vtrack->demuxer->data_size - vtrack->demuxer->data_position <
@@ -654,6 +817,7 @@ int mpeg3_toc_video(mpeg3_t *file,
 		vtrack->demuxer->data_position];
 	uint32_t code = (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | (ptr[3]);
 	ptr += 4;
+
 	while(vtrack->demuxer->data_size - vtrack->demuxer->data_position > 
 		MPEG3_VIDEO_STREAM_SIZE)
 	{
@@ -661,10 +825,11 @@ int mpeg3_toc_video(mpeg3_t *file,
 			code == MPEG3_GOP_START_CODE ||
 			code == MPEG3_PICTURE_START_CODE)
 		{
+			if(vtrack->prev_frame_offset == -1) vtrack->prev_frame_offset = vtrack->prev_offset;
 
 // Use video decoder to get repeat count and field type.  Should never hit EOF in here.
 // This rereads up to the current ptr since data_position isn't updated by
-// mpeg3_toc_video.
+// handle_video.
 			if(!mpeg3video_get_header(video, 0))
 			{
 				if(video->pict_struct == BOTTOM_FIELD ||
@@ -674,21 +839,41 @@ int mpeg3_toc_video(mpeg3_t *file,
 					int is_keyframe = (video->pict_type == I_TYPE);
 
 // Add entry for every repeat count.
-					mpeg3_append_frame(vtrack, vtrack->prev_offset, is_keyframe);
+					mpeg3_append_frame(vtrack, vtrack->prev_frame_offset, is_keyframe);
 					video->current_repeat += 100;
 					while(video->repeat_count - video->current_repeat >= 100)
 					{
-						mpeg3_append_frame(vtrack, vtrack->prev_offset, is_keyframe);
+						mpeg3_append_frame(vtrack, vtrack->prev_frame_offset, is_keyframe);
 						video->current_repeat += 100;
 					}
 
+// Shift out data from before frame
+					mpeg3demux_shift_data(vtrack->demuxer, vtrack->demuxer->data_position);
+
+// Reset pointer
 					ptr = &vtrack->demuxer->data_buffer[
 						vtrack->demuxer->data_position];
 					code = (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | (ptr[3]);
 					ptr += 4;
 
-// Shift out data from before frame
-					mpeg3demux_shift_data(vtrack->demuxer, vtrack->demuxer->data_position);
+					vtrack->prev_frame_offset = -1;
+				}
+				else
+				{
+/*
+ * // Shift out data from before frame
+ * 					vtrack->demuxer->data_position++;
+ * 					mpeg3demux_shift_data(vtrack->demuxer, vtrack->demuxer->data_position);
+ * 
+ * // Reset pointer
+ * 					ptr = &vtrack->demuxer->data_buffer[
+ * 						vtrack->demuxer->data_position];
+ * 					code = (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | (ptr[3]);
+ * 					ptr += 4;
+ * 
+ * 					vtrack->prev_frame_offset = -1;
+ */
+					break;
 				}
 			}
 			else
@@ -700,16 +885,54 @@ int mpeg3_toc_video(mpeg3_t *file,
 		else
 		{
 			vtrack->demuxer->data_position++;
-			code <<= 8;
-			code |= *ptr++;
+//			code <<= 8;
+//			code |= *ptr++;
+			ptr = &vtrack->demuxer->data_buffer[vtrack->demuxer->data_position];
+			code = (ptr[0] << 24) | (ptr[1] << 16) | (ptr[2] << 8) | (ptr[3]);  
 		}
 	}
+//printf("handle_video 20\n");
 	vtrack->demuxer->data_position -= 4;
+	mpeg3demux_shift_data(vtrack->demuxer, vtrack->demuxer->data_position);
 
 
 	return 0;
 }
 
+
+static void handle_subtitle(mpeg3_t *file)
+{
+	int i, j;
+	int got_one = 1;
+	while(got_one)
+	{
+		got_one = 0;
+
+		mpeg3_subtitle_t *subtitle = 0;
+		mpeg3_strack_t *strack = 0;
+		for(i = 0; i < mpeg3_subtitle_tracks(file); i++)
+		{
+			strack = mpeg3_get_strack(file, i);
+			if((subtitle = mpeg3_get_subtitle(strack)))
+			{
+				got_one = 1;
+				break;
+			}
+		}
+
+
+		if(subtitle)
+		{
+/* Add offset to the subtitle track */
+			mpeg3_append_subtitle_offset(strack, subtitle->offset);
+
+
+/* Remove from buffer */
+			mpeg3_pop_subtitle(strack, 0, 1);
+		}
+	}
+}
+
 int mpeg3_do_toc(mpeg3_t *file, int64_t *bytes_processed)
 {
 	int i, j, k;
@@ -718,9 +941,11 @@ int mpeg3_do_toc(mpeg3_t *file, int64_t *bytes_processed)
 
 	start_byte = mpeg3demux_tell_byte(file->demuxer);
 
+//printf("mpeg3_do_toc 1\n");
 	int result = mpeg3_read_next_packet(file->demuxer);
+//printf("mpeg3_do_toc 10\n");
 
-// Determine program and stream id for current packet.  Should be only one.
+// Determine program interleaving for current packet.
 	int program = mpeg3demux_tell_program(file->demuxer);
 
 
@@ -745,8 +970,16 @@ int mpeg3_do_toc(mpeg3_t *file, int64_t *bytes_processed)
 		int got_it = 0;
 
 
-// In a transport stream the PID's are unique for all audio and video but in 
-// other streams the PID's are shared.
+
+// Got subtitle
+		if(file->demuxer->got_subtitle)
+		{
+			handle_subtitle(file);
+		}
+
+
+// In a transport stream the audio or video is determined by the PID.
+// In other streams the data type is determined by stream ID.
 		if(file->demuxer->got_audio || 
 			file->is_transport_stream ||
 			file->is_audio_stream)
@@ -756,9 +989,13 @@ int mpeg3_do_toc(mpeg3_t *file, int64_t *bytes_processed)
 				mpeg3_atrack_t *atrack = file->atrack[i];
 				if(custom_id == atrack->pid)
 				{
-//printf("mpeg3_do_toc 2 %x\n", atrack->pid);
+/*
+ * printf("mpeg3_do_toc 2 offset=0x%llx pid=0x%x\n", 
+ * start_byte, 
+ * atrack->pid);
+ */
 // Update an audio track
-					mpeg3_toc_audio(file, i);
+					handle_audio(file, i);
 					atrack->prev_offset = start_byte;
 					got_it = 1;
 					break;
@@ -790,7 +1027,7 @@ int mpeg3_do_toc(mpeg3_t *file, int64_t *bytes_processed)
 					file->total_astreams++;
 // Make the first offset correspond to the start of the first packet.
 					mpeg3_append_samples(atrack, start_byte);
-					mpeg3_toc_audio(file, file->total_astreams - 1);
+					handle_audio(file, file->total_astreams - 1);
 					atrack->prev_offset = start_byte;
 				}
 			}
@@ -809,7 +1046,7 @@ int mpeg3_do_toc(mpeg3_t *file, int64_t *bytes_processed)
 				if(vtrack->pid == custom_id)
 				{
 // Update a video track
-					mpeg3_toc_video(file, vtrack);
+					handle_video(file, vtrack);
 					vtrack->prev_offset = start_byte;
 					got_it = 1;
 					break;
@@ -835,7 +1072,7 @@ int mpeg3_do_toc(mpeg3_t *file, int64_t *bytes_processed)
 					file->total_vstreams++;
 // Create table entry for frame 0
 					mpeg3_append_frame(vtrack, start_byte, 1);
-					mpeg3_toc_video(file, vtrack);
+					handle_video(file, vtrack);
 					vtrack->prev_offset = start_byte;
 				}
 			}
@@ -852,54 +1089,12 @@ int mpeg3_do_toc(mpeg3_t *file, int64_t *bytes_processed)
 
 
 
-#define PUT_INT32(x) \
-{ \
-	if(MPEG3_LITTLE_ENDIAN) \
-	{ \
-		fputc(((unsigned char*)&x)[3], file->toc_fd); \
-		fputc(((unsigned char*)&x)[2], file->toc_fd); \
-		fputc(((unsigned char*)&x)[1], file->toc_fd); \
-		fputc(((unsigned char*)&x)[0], file->toc_fd); \
-	} \
-	else \
-	{ \
-		fputc(((unsigned char*)&x)[0], file->toc_fd); \
-		fputc(((unsigned char*)&x)[1], file->toc_fd); \
-		fputc(((unsigned char*)&x)[2], file->toc_fd); \
-		fputc(((unsigned char*)&x)[3], file->toc_fd); \
-	} \
-}
-
-
-
-
-#define PUT_INT64(x) \
-{ \
-	if(MPEG3_LITTLE_ENDIAN) \
-	{ \
-		fputc(((unsigned char*)&x)[7], file->toc_fd); \
-		fputc(((unsigned char*)&x)[6], file->toc_fd); \
-		fputc(((unsigned char*)&x)[5], file->toc_fd); \
-		fputc(((unsigned char*)&x)[4], file->toc_fd); \
-		fputc(((unsigned char*)&x)[3], file->toc_fd); \
-		fputc(((unsigned char*)&x)[2], file->toc_fd); \
-		fputc(((unsigned char*)&x)[1], file->toc_fd); \
-		fputc(((unsigned char*)&x)[0], file->toc_fd); \
-	} \
-	else \
-	{ \
-		fwrite(&x, 1, 8, file->toc_fd); \
-	} \
-}
-
-
-
 
 
 void mpeg3_stop_toc(mpeg3_t *file)
 {
 // Create final chunk for audio tracks to count the last samples.
-	int i, j;
+	int i, j, k;
 	for(i = 0; i < file->total_astreams; i++)
 	{
 		mpeg3_atrack_t *atrack = file->atrack[i];
@@ -984,43 +1179,51 @@ void mpeg3_stop_toc(mpeg3_t *file)
 	fputc(' ', file->toc_fd);
 
 // Write version
-	fputc(MPEG3_TOC_VERSION, file->toc_fd);
+	PUT_INT32(MPEG3_TOC_VERSION);
 
 // Write stream type
 	if(file->is_program_stream)
 	{
-		fputc(FILE_TYPE_PROGRAM, file->toc_fd);
+		PUT_INT32(FILE_TYPE_PROGRAM);
 	}
 	else
 	if(file->is_transport_stream)
 	{
-		fputc(FILE_TYPE_TRANSPORT, file->toc_fd);
+		PUT_INT32(FILE_TYPE_TRANSPORT);
 	}
 	else
 	if(file->is_audio_stream)
 	{
-		fputc(FILE_TYPE_AUDIO, file->toc_fd);
+		PUT_INT32(FILE_TYPE_AUDIO);
 	}
 	else
 	if(file->is_video_stream)
 	{
-		fputc(FILE_TYPE_VIDEO, file->toc_fd);
+		PUT_INT32(FILE_TYPE_VIDEO);
 	}
 
+
+// Store file information
+	PUT_INT32(FILE_INFO);
+	fprintf(file->toc_fd, file->fs->path);
+	for(j = strlen(file->fs->path); j < MPEG3_STRLEN; j++)
+			fputc(0, file->toc_fd);
+	PUT_INT64(file->source_date);
+
 // Write stream ID's
 // Only program and transport streams have these
 	for(i = 0; i < MPEG3_MAX_STREAMS; i++)
 	{
 		if(file->demuxer->astream_table[i])
 		{
-			fputc(STREAM_AUDIO, file->toc_fd);
+			PUT_INT32(STREAM_AUDIO);
 			PUT_INT32(i);
 			PUT_INT32(file->demuxer->astream_table[i]);
 		}
 
 		if(file->demuxer->vstream_table[i])
 		{
-			fputc(STREAM_VIDEO, file->toc_fd);
+			PUT_INT32(STREAM_VIDEO);
 			PUT_INT32(i);
 			PUT_INT32(file->demuxer->vstream_table[i]);
 		}
@@ -1031,16 +1234,21 @@ void mpeg3_stop_toc(mpeg3_t *file)
 	{
 		mpeg3_title_t *title = file->demuxer->titles[i];
 // Path
-		fputc(TITLE_PATH, file->toc_fd);
+		PUT_INT32(TITLE_PATH);
+
 		fprintf(file->toc_fd, title->fs->path);
-		fputc(0, file->toc_fd);
+
+// Pad path with 0
+		for(j = strlen(title->fs->path); j < MPEG3_STRLEN; j++)
+			fputc(0, file->toc_fd);
 // Total bytes
 		PUT_INT64(title->total_bytes);
-// Byte offsets of cells
+// Total cells in title
 		PUT_INT32(file->demuxer->titles[i]->cell_table_size);
 		for(j = 0; j < title->cell_table_size; j++)
 		{
 			mpeg3_cell_t *cell = &title->cell_table[j];
+//printf("%x: %llx-%llx %llx-%llx %d\n", ftell(file->toc_fd), cell->title_start, cell->title_end, cell->program_start, cell->program_end, cell->program);
 			PUT_INT64(cell->title_start);
 			PUT_INT64(cell->title_end);
 			PUT_INT64(cell->program_start);
@@ -1050,13 +1258,9 @@ void mpeg3_stop_toc(mpeg3_t *file)
 	}
 
 
-
-	fputc(ATRACK_COUNT, file->toc_fd);
+	PUT_INT32(ATRACK_COUNT);
 	PUT_INT32(file->total_astreams);
 
-	fputc(VTRACK_COUNT, file->toc_fd);
-	PUT_INT32(file->total_vstreams);
-
 // Audio streams
 	for(j = 0; j < file->total_astreams; j++)
 	{
@@ -1064,16 +1268,42 @@ void mpeg3_stop_toc(mpeg3_t *file)
 		PUT_INT64(atrack->audio_eof);
 		PUT_INT32(atrack->channels);
 		PUT_INT32(atrack->total_sample_offsets);
+// Total samples
+		PUT_INT64(atrack->current_position);
+// Sample offsets
 		for(i = 0; i < atrack->total_sample_offsets; i++)
 		{
 			PUT_INT64(atrack->sample_offsets[i]);
 		}
+
+// Index
+		mpeg3_index_t *index = file->indexes[j];
+		if(index->index_data)
+		{
+			PUT_INT32(index->index_size);
+			PUT_INT32(index->index_zoom);
+			for(k = 0; k < atrack->channels; k++)
+			{
+				fwrite(index->index_data[k], 
+					sizeof(float) * 2, 
+					index->index_size,
+					file->toc_fd);
+			}
+		}
+		else
+		{
+			PUT_INT32(0);
+			PUT_INT32(1);
+		}
 	}
 
 
 
 
 
+	PUT_INT32(VTRACK_COUNT);
+	PUT_INT32(file->total_vstreams);
+
 // Video streams
 	for(j = 0; j < file->total_vstreams; j++)
 	{
@@ -1093,27 +1323,29 @@ void mpeg3_stop_toc(mpeg3_t *file)
 	}
 
 
-// Audio indexes
-// Write indexes
-	for(i = 0; i < file->total_astreams; i++)
+
+
+	PUT_INT32(STRACK_COUNT);
+	PUT_INT32(file->total_sstreams);
+
+// Subtitle tracks
+	for(i = 0; i < file->total_sstreams; i++)
 	{
-		mpeg3_atrack_t *atrack = file->atrack[i];
-		mpeg3_index_t *index = file->indexes[i];
-		if(index->index_data)
+		mpeg3_strack_t *strack = file->strack[i];
+		PUT_INT32(strack->id);
+		PUT_INT32(strack->total_offsets);
+		for(j = 0; j < strack->total_offsets; j++)
 		{
-			PUT_INT32(index->index_size);
-			PUT_INT32(index->index_zoom);
-			for(j = 0; j < atrack->channels; j++)
-			{
-				fwrite(index->index_data[j], 
-					sizeof(float) * 2, 
-					index->index_size,
-					file->toc_fd);
-			}
+			PUT_INT64(strack->offsets[j]);
 		}
 	}
 
 
+	PUT_INT32(IFO_PALETTE);
+	for(i = 0; i < 16 * 4; i++)
+	{
+		fputc(file->palette[i], file->toc_fd);
+	}
 
 
 	fclose(file->toc_fd);
@@ -1167,7 +1399,24 @@ int mpeg3_has_toc(mpeg3_t *file)
 	return 0;
 }
 
+char* mpeg3_title_path(mpeg3_t *file, int number)
+{
+	if(number < file->demuxer->total_titles)
+	{
+		return file->demuxer->titles[number]->fs->path;
+	}
+	return 0;
+}
 
+int64_t mpeg3_get_source_date(mpeg3_t *file)
+{
+	return file->source_date;
+}
 
-
-
+int64_t mpeg3_calculate_source_date(char *path)
+{
+	struct stat64 ostat;
+	bzero(&ostat, sizeof(struct stat64));
+	stat64(path, &ostat);
+	return ostat.st_mtime;
+}
