@@ -12,6 +12,42 @@ void mpeg3video_toc_error()
 		"to generate a table of contents and load the table of contents instead.\n");
 }
 
+int mpeg3video_drop_frames(mpeg3video_t *video, long frames, int cache_it)
+{
+	int result = 0;
+	long frame_number = video->framenum + frames;
+	mpeg3_vtrack_t *track = video->track;
+/* first 3 frames are garbage as proven by printfs in the read_frame calls */
+	int drop_count = 3;
+
+/* Read the selected number of frames and skip b-frames */
+	while(!result && frame_number > video->framenum)
+	{
+		if(cache_it)
+		{
+			result = mpeg3video_read_frame_backend(video, 0);
+        	if(video->output_src[0] && drop_count--)
+        	{
+				mpeg3_cache_put_frame(track->frame_cache,
+					video->framenum - 1,
+					video->output_src[0],
+					video->output_src[1],
+					video->output_src[2],
+					video->coded_picture_width * video->coded_picture_height,
+					video->chrom_width * video->chrom_height,
+					video->chrom_width * video->chrom_height);
+//printf("mpeg3video_drop_frames 1 %d\n", video->framenum);
+        	}
+		}
+		else
+		{
+			result = mpeg3video_read_frame_backend(video, frame_number - video->framenum);
+		}
+	}
+
+	return result;
+}
+
 unsigned int mpeg3bits_next_startcode(mpeg3_bits_t* stream)
 {
 /* Perform forwards search */
@@ -191,6 +227,11 @@ int mpeg3video_seek(mpeg3video_t *video)
 		video->byte_seek = -1;
 		mpeg3demux_seek_byte(demuxer, byte);
 
+
+// Clear subtitles
+		mpeg3_reset_subtitles(file);
+
+
 // Rewind 2 I-frames
 		if(byte > 0)
 		{
@@ -259,6 +300,10 @@ int mpeg3video_seek(mpeg3video_t *video)
 /* Seek to a frame */
 	if(video->frame_seek >= 0)
 	{
+// Clear subtitles
+		mpeg3_reset_subtitles(file);
+
+
 		frame_number = video->frame_seek;
 		video->frame_seek = -1;
 		if(frame_number < 0) frame_number = 0;
@@ -266,9 +311,13 @@ int mpeg3video_seek(mpeg3video_t *video)
 
 //printf("mpeg3video_seek 1 %ld %ld\n", frame_number, video->framenum);
 
-/* Seek to I frame in table of contents */
+/* Seek to I frame in table of contents. */
+/* Determine time between seek position and previous subtitle. */
+/* Subtract time difference from subtitle display time. */
 		if(track->frame_offsets)
 		{
+			mpeg3_reset_cache(track->frame_cache);
+
 			if((frame_number < video->framenum || 
 				frame_number - video->framenum > MPEG3_SEEK_THRESHOLD))
 			{
@@ -303,7 +352,9 @@ int mpeg3video_seek(mpeg3video_t *video)
 					
 
 						video->repeat_count = 0;
-						mpeg3video_drop_frames(video, frame_number - video->framenum);
+
+// Read up to current frame
+						mpeg3video_drop_frames(video, frame_number - video->framenum, 1);
 						break;
 					}
 				}
@@ -311,138 +362,17 @@ int mpeg3video_seek(mpeg3video_t *video)
 			else
 			{
 				video->repeat_count = 0;
-				mpeg3video_drop_frames(video, frame_number - video->framenum);
+				mpeg3video_drop_frames(video, frame_number - video->framenum, 0);
 			}
 		}
 		else
-/* Discontinue support of seeking without table of contents */
+/* No support for seeking without table of contents */
 		{
 			mpeg3video_toc_error();
 		}
 
 
 
-
-
-
-
-
-
-
-
-
-
-#if 0
-/* Seek to start of file */
-		if(frame_number < 16)
-		{
-			video->repeat_count = video->current_repeat = 0;
-			mpeg3bits_seek_start(vstream);
-			video->framenum = 0;
-			result = mpeg3video_drop_frames(video, 
-				frame_number - video->framenum);
-		}
-		else
-		{
-/* Seek to an I frame. */
-//printf(__FUNCTION__ " frame_number=%d video->framenum=%d\n", frame_number, video->framenum);
-			if((frame_number < video->framenum || 
-				frame_number - video->framenum > MPEG3_SEEK_THRESHOLD))
-			{
-
-
-
-/* Elementary stream.  Estimate frame position from total bytes. */
-				if(file->is_video_stream)
-				{
-					int64_t byte = (int64_t)((double)(mpeg3demux_movie_size(demuxer) / 
-						track->total_frames) * 
-						frame_number);
-					long minimum = 65535;
-					int done = 0;
-
-/* Get GOP just before frame */
-					do
-					{
-						result = mpeg3bits_seek_byte(vstream, byte);
-						mpeg3bits_start_reverse(vstream);
-
-						if(!result) result = mpeg3video_prev_code(vstream, MPEG3_GOP_START_CODE);
-						mpeg3bits_start_forward(vstream);
-						mpeg3bits_getbits(vstream, 8);
-						if(!result) result = mpeg3video_getgophdr(video);
-						this_gop_start = mpeg3video_goptimecode_to_frame(video);
-
-//printf("wanted %ld guessed %ld byte %ld result %d\n", frame_number, this_gop_start, byte, result);
-						if(labs(this_gop_start - frame_number) >= labs(minimum)) 
-							done = 1;
-						else
-						{
-							minimum = this_gop_start - frame_number;
-							byte += (long)((float)(frame_number - this_gop_start) * 
-								(double)(mpeg3demux_movie_size(demuxer) / 
-								track->total_frames));
-							if(byte < 0) byte = 0;
-						}
-					}while(!result && !done);
-
-//printf("wanted %d guessed %d\n", frame_number, this_gop_start);
-					if(!result)
-					{
-						video->framenum = this_gop_start;
-						result = mpeg3video_drop_frames(video, frame_number - video->framenum);
-					}
-				}
-				else
-
-
-
-
-/* System stream */
-				{
-					mpeg3bits_seek_time(vstream, (double)frame_number / video->frame_rate);
-
-					byte = mpeg3bits_tell(vstream);
-					mpeg3bits_start_reverse(vstream);
-					mpeg3video_prev_code(vstream, MPEG3_GOP_START_CODE);
-					mpeg3bits_getbits_reverse(vstream, 32);
-					mpeg3bits_start_forward(vstream);
-
-					while(!result && mpeg3bits_tell(vstream) < byte)
-					{
-						result = mpeg3video_read_frame_backend(video, 0);
-						if(match_refframes)
-							mpeg3video_match_refframes(video);
-
-						match_refframes = 0;
-					}
-//printf("seek system 3 %f\n", (double)frame_number / video->frame_rate);
-				}
-
-				video->framenum = frame_number;
-			}
-			else
-// Drop frames
-			{
-				mpeg3video_drop_frames(video, frame_number - video->framenum);
-			}
-		}
-#endif
-
-
-
-
-
-
-
-
-
-
-
-
-
-
-
 		mpeg3demux_reset_pts(demuxer);
 	}
 
@@ -499,19 +429,3 @@ int mpeg3video_previous_frame(mpeg3video_t *video)
 	video->repeat_count = 0;
 	return 0;
 }
-
-int mpeg3video_drop_frames(mpeg3video_t *video, long frames)
-{
-	int result = 0;
-	long frame_number = video->framenum + frames;
-
-//printf("mpeg3video_drop_frames 1 %d %d\n", frame_number, video->framenum);
-/* Read the selected number of frames and skip b-frames */
-	while(!result && frame_number > video->framenum)
-	{
-		result = mpeg3video_read_frame_backend(video, frame_number - video->framenum);
-	}
-//printf("mpeg3video_drop_frames 100\n");
-
-	return result;
-}
